import logging
from queue import Queue as ThreadQueue

from .error import ChannelClosed
from .newio_broker import NewioBroker
from .thread_broker import ThreadBroker

LOG = logging.getLogger(__name__)


class ThreadChannel:
    '''Message channel for communicating between threads and newio tasks'''

    def __init__(self, maxsize=0):
        self._queue = ThreadQueue(maxsize)
        self._thread_broker = ThreadBroker(self._queue)
        self._newio_broker = NewioBroker(self._queue)
        self._is_closed = False

    def __repr__(self):
        return f'<ThreadChannel at {hex(id(self))}>'

    async def __aenter__(self):
        if self._is_closed:
            raise RuntimeError('Channel already closed')
        await self._thread_broker.start()
        await self._newio_broker.start()
        return self

    async def __aexit__(self, *exc_info):
        self._is_closed = True
        await self._thread_broker.shutdown()
        await self._newio_broker.shutdown()

    async def __aiter__(self):
        while True:
            try:
                yield (await self.recv())
            except ChannelClosed:
                break

    def thread_iter(self):
        while True:
            try:
                yield self.thread_recv()
            except ChannelClosed:
                break

    def thread_send(self, item):
        '''send in thread'''
        if self._is_closed:
            raise ChannelClosed()
        self._thread_broker.send(item)
        self._newio_broker.notify_recv()

    def thread_recv(self):
        '''recv in thread'''
        item = self._thread_broker.recv()
        self._newio_broker.notify_send()
        return item

    async def send(self, item):
        '''send in newio task'''
        if self._is_closed:
            raise ChannelClosed()
        await self._newio_broker.send(item)
        self._thread_broker.notify_recv()

    async def recv(self):
        '''recv in newio task'''
        item = await self._newio_broker.recv()
        self._thread_broker.notify_send()
        return item



'''Newio common API for users'''
from . import syscall
from .syscall import TaskCanceled, TaskTimeout, Timer, Task, Lounge

__all__ = (
    'TaskCanceled', 'TaskTimeout', 'Task', 'Timer', 'Lounge',
    'wait_read', 'wait_write', 'sleep', 'spawn', 'current_task',
    'run_in_thread', 'run_in_process', 'run_in_asyncio',
    'timeout_after', 'open_nursery',
)

wait_read = syscall.nio_wait_read
wait_write = syscall.nio_wait_write
run_in_thread = syscall.nio_run_in_thread
run_in_process = syscall.nio_run_in_process
run_in_asyncio = syscall.nio_run_in_asyncio
sleep = syscall.nio_sleep
spawn = syscall.nio_spawn
current_task = syscall.nio_current_task


class timeout_after:
    '''Async context manager for task timeout

    Usage:

        async with timeout_after(3) as is_timeout:
            await sleep(5)
        if is_timeout:
            # task timeout
    '''

    def __init__(self, seconds: float):
        self._seconds = seconds
        self._timer = None

    def __bool__(self):
        return self._timer is not None and self._timer.is_expired

    async def __aenter__(self):
        self._timer = await syscall.nio_timeout_after(self._seconds)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await syscall.nio_unset_timer(self._timer)
        should_suppress = (
            exc_value is not None and
            isinstance(exc_value, TaskTimeout) and
            exc_value.timer is self._timer
        )
        return should_suppress


def open_nursery():
    return _Nursery()


class _Nursery:
    '''Nursery is manager of tasks, it will take care of it spawned tasks.

    All tasks spawned by the nursery are ensure stoped after nursery exited.
    When nursery exiting, it will join spawned tasks, if join failed
    it will cancel spawned tasks.
    '''

    def __init__(self):
        self._tasks = []
        self._is_closed = False

    async def spawn(self, coro):
        '''Spawn task in the nursery'''
        if self._is_closed:
            raise RuntimeError('nursery already closed')
        task = await spawn(coro)
        self._tasks.append(task)
        return task

    async def _join(self):
        for task in self._tasks:
            if task.is_alive:
                await task.join()

    async def _cancel(self):
        for task in self._tasks:
            if task.is_alive:
                await task.cancel()

    async def __aenter__(self):
        if self._is_closed:
            raise RuntimeError('nursery already closed')
        return self

    async def __aexit__(self, *exc_info):
        self._is_closed = True
        try:
            await self._join()
        finally:
            await self._cancel()
            await self._join()



import logging
from queue import Full, Empty
from threading import Lock as ThreadLock

from newio import spawn
from newio.sync import Condition
from newio.socket import socketpair
from .error import ChannelClosed

LOG = logging.getLogger(__name__)


class NewioBroker:
    def __init__(self, queue):
        self._queue = queue
        self._notify_send, self._wakeup_send = socketpair()
        self._notify_recv, self._wakeup_recv = socketpair()
        self._send_cond = Condition()
        self._recv_cond = Condition()
        self._is_closed = False
        self._notify_lock = ThreadLock()
        self._consumer_broker_task = None
        self._producer_broker_task = None

    async def start(self):
        self._consumer_broker_task = await spawn(self._consumer_broker())
        self._producer_broker_task = await spawn(self._producer_broker())

    async def shutdown(self, wait=True):
        LOG.debug('broker exiting')
        self._is_closed = True
        with self._notify_lock:
            self._notify_send.socket.sendall(b'x')
            self._notify_recv.socket.sendall(b'x')
        await self._notify_send.close()
        await self._notify_recv.close()
        await self._send_cond.notify_all()
        await self._recv_cond.notify_all()
        if wait:
            await self._consumer_broker_task.join()
            await self._producer_broker_task.join()
            await self._wakeup_send.close()
            await self._wakeup_recv.close()

    def notify_send(self):
        if self._is_closed:
            return
        with self._notify_lock:
            self._notify_send.socket.sendall(b'1')

    def notify_recv(self):
        if self._is_closed:
            return
        with self._notify_lock:
            self._notify_recv.socket.sendall(b'1')

    async def send(self, item):
        if self._is_closed:
            raise ChannelClosed()
        while True:
            try:
                self._queue.put_nowait(item)
            except Full:
                await self._send_cond.wait()
                if self._is_closed:
                    raise ChannelClosed() from None
            else:
                break
        await self._recv_cond.notify()

    async def recv(self):
        while True:
            try:
                item = self._queue.get_nowait()
            except Empty:
                if self._is_closed:
                    raise ChannelClosed() from None
                await self._recv_cond.wait()
            else:
                break
        await self._send_cond.notify()
        return item

    async def _consumer_broker(self):
        LOG.debug('consumer broker started')
        while True:
            nbytes = await self._wakeup_recv.recv(1)
            if not nbytes or nbytes == b'x':
                LOG.debug('consumer broker exiting')
                break
            await self._recv_cond.notify()

    async def _producer_broker(self):
        LOG.debug('producer broker started')
        while True:
            nbytes = await self._wakeup_send.recv(1)
            if not nbytes or nbytes == b'x':
                LOG.debug('producer broker exiting')
                break
            await self._send_cond.notify()



import logging
from queue import Empty, Full

from .error import ChannelClosed

LOG = logging.getLogger(__name__)


class ThreadBroker:
    def __init__(self, queue):
        self._queue = queue
        self._blocking_timeout = 0.1
        self._is_closed = False

    async def start(self):
        '''do nothing'''

    async def shutdown(self, wait=True):
        LOG.debug('broker exiting')
        self._is_closed = True

    def notify_send(self):
        '''do nothing'''

    def notify_recv(self):
        '''do nothing'''

    def send(self, item):
        if self._is_closed:
            raise ChannelClosed()
        while True:
            try:
                self._queue.put(item, timeout=self._blocking_timeout)
            except Full:
                if self._is_closed:
                    raise ChannelClosed() from None
            else:
                break

    def recv(self):
        if self._is_closed:
            try:
                return self._queue.get_nowait()
            except Empty:
                raise ChannelClosed() from None
        while True:
            try:
                item = self._queue.get(timeout=self._blocking_timeout)
            except Empty:
                if self._is_closed:
                    raise ChannelClosed() from None
            else:
                break
        return item



import logging
import asyncio
from queue import Empty, Full
from threading import Lock as ThreadLock
from socket import socketpair

from newio.api import spawn, run_in_asyncio
from .error import ChannelClosed

LOG = logging.getLogger(__name__)


async def cond_wait(cond):
    await cond.acquire()
    try:
        await cond.wait()
    finally:
        cond.release()


async def cond_notify(cond):
    await cond.acquire()
    try:
        cond.notify()
    finally:
        cond.release()


async def cond_notify_all(cond):
    await cond.acquire()
    try:
        cond.notify_all()
    finally:
        cond.release()


class AsyncioBroker:
    def __init__(self, queue):
        self._queue = queue
        self._notify_send, self._wakeup_send = socketpair()
        self._notify_recv, self._wakeup_recv = socketpair()
        self._notify_send.setblocking(False)
        self._wakeup_send.setblocking(False)
        self._notify_recv.setblocking(False)
        self._wakeup_recv.setblocking(False)
        self._send_cond = None
        self._recv_cond = None
        self._is_closed = False
        self._notify_lock = ThreadLock()
        self._consumer_broker_task = None
        self._producer_broker_task = None

    async def _init_cond(self):

        async def _init():
            self._send_cond = asyncio.Condition()
            self._recv_cond = asyncio.Condition()

        await run_in_asyncio(_init())

    async def _notify_all_cond(self):

        async def _notify_all():
            await cond_notify_all(self._send_cond)
            await cond_notify_all(self._recv_cond)

        await run_in_asyncio(_notify_all())

    async def start(self):
        await self._init_cond()
        self._consumer_broker_task = await spawn(self._consumer_broker())
        self._producer_broker_task = await spawn(self._producer_broker())

    async def shutdown(self, wait=True):
        LOG.debug('broker exiting')
        self._is_closed = True
        with self._notify_lock:
            self._notify_send.sendall(b'x')
            self._notify_recv.sendall(b'x')
        self._notify_send.close()
        self._notify_recv.close()
        await self._notify_all_cond()
        if wait:
            await self._consumer_broker_task.join()
            await self._producer_broker_task.join()
            self._wakeup_send.close()
            self._wakeup_recv.close()

    def notify_send(self):
        if self._is_closed:
            return
        with self._notify_lock:
            self._notify_send.sendall(b'1')

    def notify_recv(self):
        if self._is_closed:
            return
        with self._notify_lock:
            self._notify_recv.sendall(b'1')

    async def send(self, item):
        if self._is_closed:
            raise ChannelClosed()
        while True:
            try:
                self._queue.put_nowait(item)
            except Full:
                await cond_wait(self._send_cond)
                if self._is_closed:
                    raise ChannelClosed() from None
            else:
                break
        await cond_notify(self._recv_cond)

    async def recv(self):
        while True:
            try:
                item = self._queue.get_nowait()
            except Empty:
                if self._is_closed:
                    raise ChannelClosed() from None
                await cond_wait(self._recv_cond)
            else:
                break
        await cond_notify(self._send_cond)
        return item

    async def _consumer_broker(self):

        async def _broker():
            LOG.debug('consumer broker started')
            loop = asyncio.get_event_loop()
            while True:
                nbytes = await loop.sock_recv(self._wakeup_recv, 1)
                if not nbytes or nbytes == b'x':
                    LOG.debug('consumer broker exiting')
                    break
                await cond_notify(self._recv_cond)

        await run_in_asyncio(_broker())

    async def _producer_broker(self):

        async def _broker():
            LOG.debug('producer broker started')
            loop = asyncio.get_event_loop()
            while True:
                nbytes = await loop.sock_recv(self._wakeup_send, 1)
                if not nbytes or nbytes == b'x':
                    LOG.debug('producer broker exiting')
                    break
                await cond_notify(self._send_cond)

        await run_in_asyncio(_broker())



import _newio
from setuptools import setup

version = _newio.__version__

setup(
    name='newio',
    version=version,
    description='Newio',
    url='https://github.com/guyskk/newio',
    author='guyskk',
    author_email='guyskk@qq.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    packages=_newio.find_packages('newio'),
    package_data=_newio.package_data,
    python_requires='>=3.6',
    install_requires=[],
    extras_require={
        'kernel': [f'newio-kernel=={version}'],
    },
)



from pathlib import Path
from setuptools import find_packages as _find_packages

_version_file = Path(__file__).parent / 'version.txt'

__version__ = _version_file.read_text().strip()


def parse_requires(name):
    with open(name) as f:
        return f.read().splitlines()


package_data = {
    '_newio': ['version.txt']
}


def find_packages(target):
    target = target.replace('-', '_')
    packages = ['_newio', target]
    for name in _find_packages(target):
        packages.append(f'{target}.{name}')
    return packages
